import numpy as np
import matplotlib.pyplot as plt

def sigma(x):
    return np.tanh(x)

def layer(x, K_mat, b, h):
    return x + h * sigma( np.matmul(K_mat,x) +b )

def feed_forward(x0, K_mat, b, h, N):
    x = x0.reshape(1, -1)
    for i in range(N):
        x = np.vstack( (x, layer(x[-1,:], K_mat, b, h)) )
    return x

x0_1 = np.array([1., 0.4])
x0_2 = np.array([0.99, 0.46])
x0_3 = np.array([0.95, 0.5])
N = 20
h = 0.1

#=== Stable Net ===
K_mat = np.array([[0., 1.], [-1., 0.]])
b = np.array([0., 0.])
x_stable = []
for x0 in [x0_1, x0_2, x0_3]:
    x_stable.append( feed_forward(x0, K_mat, b, h, N) )

#=== Contractive Net ===
K_mat = np.array([[-2., 0.], [2., -2.]])
b = np.array([0., 0.])
x_contractive = []
for x0 in [x0_1, x0_2, x0_3]:
    x_contractive.append( feed_forward(x0, K_mat, b, h, N) )

#=== Expansive Net ===
K_mat = np.array([[2., -2.], [0., 2.]])
b = np.array([0., 0.])
x_expansive = []
for x0 in [x0_1, x0_2, x0_3]:
    x_expansive.append( feed_forward(x0, K_mat, b, h, N) )

#=== PLOTTING ===

for trajectories in zip([x_stable, x_contractive, x_expansive],
    ["Stable NN", "Contractive NN", "Expansive NN"]):

    fig, ax = plt.subplots()
    for x in trajectories[0]:
        ax.scatter(x[:,0], x[:,1])

    plt.title(trajectories[1], y=1.08)
    lim = 1.2 * np.max(trajectories[0])
    ax.set(xlim=(-lim, lim), ylim=(-lim, lim), aspect="equal")

    # set up "classical" cartesian coordinate axes
    ax.spines["bottom"].set_position("zero")
    ax.spines["left"].set_position("zero")
    ax.spines["right"].set_visible(False)
    ax.spines["top"].set_visible(False)
    ax.set_xlabel('x', labelpad=-22, x=1.03)
    ax.set_ylabel('y', labelpad=-31, y=1.02, rotation=0)

    # create ticks, most importantly to remove the ticks at (x,y) = (0,0)
    num_ticks = 2
    ticks = np.linspace(-np.floor(lim), np.floor(lim), 2*num_ticks+1)
    ax.set_xticks(ticks[ticks != 0])
    ax.set_yticks(ticks[ticks != 0])

    plt.grid()

plt.show()
